@@ -1,5 +1,5 @@
 /*
- *  Copyright 2003-2007 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2003-2007 Adrian Thurston <thurston@complang.org>
  */
 
 /*  This file is part of Kelbt.
@@ -186,6 +186,7 @@ void ParseData::writeTypes()
 		"	int type;\n"
 		"	int reduction;\n" /* Should be able to extract this from the FSM. */
 		"	int state;\n"
+		"	int causeReduce;\n"
 		"	union " << UserData() << " user;\n";
 
 	if ( elementBlock != 0 )
@@ -244,7 +245,7 @@ void ParseData::writeInit()
 				"malloc( sizeof(struct " << Block() << ") );\n"
 		"	" << block() << "->next = 0;\n"
 		"	" << freshEl() << " = " << block() << "->data;\n"
-		"	#ifdef LOG_ACTIONS\n"
+		"	#ifdef KELBT_LOG_ACTIONS\n"
 		"	cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
 		"	#endif\n"
 		"	" << stackTop() << " = " << freshEl() << ";\n"
@@ -252,6 +253,7 @@ void ParseData::writeInit()
 		"	" << stackTop() << "->state = -1;\n"
 		"	" << stackTop() << "->next = 0;\n"
 		"	" << stackTop() << "->child = 0;\n"
+		"	" << stackTop() << "->causeReduce = 0;\n"
 		"	" << freshPos() << " = 1;\n"
 		"	" << lastFinal() << " = " << stackTop() << ";\n"
 		"	" << numRetry() << " = 0;\n"
@@ -450,7 +452,7 @@ void ParseData::writeData()
 	}
 	out << "\n};\n\n";
 
-	out << "char *" << prodNames() << "[] = {\n";
+	out << "const char *" << prodNames() << "[] = {\n";
 	for ( DefList::Iter prod = prodList; prod.lte(); prod++ ) {
 		out << "\t\"" << prod->data << "\"";
 		if ( ! prod.last() )
@@ -459,7 +461,7 @@ void ParseData::writeData()
 	}
 	out << "};\n\n";
 
-	out << "char *" << lelNames() << "[] = {\n";
+	out << "const char *" << lelNames() << "[] = {\n";
 	for ( int i = 0; i < nextSymbolId; i++ ) {
 		KlangEl *lel = langElIndex[i];
 		if ( lel != 0 )
@@ -670,6 +672,32 @@ void ParseData::writeFinishBlocks()
 
 }
 
+ostream &ParseData::ALLOCATE_EL( const char *name )
+{
+	out <<
+		"	if ( " << pool() << " == 0 ) {\n"
+		"		if ( " << freshPos() << " == " << FRESH_BLOCK << " ) {\n"
+		"			struct " << Block() << "* newBlock  = (struct " << Block() << "*) "
+							"malloc( sizeof(struct " << Block() << ") );\n"
+		"			newBlock->next = " << block() << ";\n"
+		"			" << block() << " = newBlock;\n"
+		"			" << freshEl() << " = newBlock->data;\n"
+		"			#ifdef KELBT_LOG_ACTIONS\n"
+		"			cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
+		"			#endif\n"
+		"			" << freshPos() << " = 0;\n"
+		"		}\n"
+		"		" << name << " = " << freshEl() << " + " << freshPos() << "++;\n"
+		"	}\n"
+		"	else {\n"
+		"		" << name << " = " << pool() << ";\n"
+		"		" << pool() << " = " << pool() << "->next;\n"
+		"	}\n"
+		"	" << numNodes() << " += 1;\n";
+
+	return out;
+}
+
 void ParseData::writeExec()
 {
 	out <<
@@ -679,51 +707,44 @@ void ParseData::writeExec()
 		"	unsigned int *action;\n"
 		"	int rhsLen;\n"
 		"	struct " << LangEl() << " *rhs[32];\n"
-		"	struct " << LangEl() << " *lel;\n"
-		"	struct " << LangEl() << " *input;\n"
+		"	struct " << LangEl() << " *lel = 0;\n"
+		"	struct " << LangEl() << " *input = 0;\n"
+		"	struct " << LangEl() << " *queue = 0;\n"
 		"	char induceReject;\n"
 		"\n"
 		"	if ( " << curs() << " < 0 )\n"
 		"		return 0;\n"
+		"\n";
+		ALLOCATE_EL( "queue" ) <<
 		"\n"
-		"	if ( " << pool() << " == 0 ) {\n"
-		"		if ( " << freshPos() << " == " << FRESH_BLOCK << " ) {\n"
-		"			struct " << Block() << "* newBlock  = (struct " << Block() << "*) "
-							"malloc( sizeof(struct " << Block() << ") );\n"
-		"			newBlock->next = " << block() << ";\n"
-		"			" << block() << " = newBlock;\n"
-		"			" << freshEl() << " = newBlock->data;\n"
-		"			#ifdef LOG_ACTIONS\n"
-		"			cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
-		"			#endif\n"
-		"			" << freshPos() << " = 0;\n"
-		"		}\n"
-		"		input = " << freshEl() << " + " << freshPos() << "++;\n"
-		"	}\n"
-		"	else {\n"
-		"		input = " << pool() << ";\n"
-		"		" << pool() << " = " << pool() << "->next;\n"
-		"	}\n"
-		"	" << numNodes() << " += 1;\n"
-		"	input->type = type;\n";
+		"	queue->type = type;\n";
 
 	if ( tokenIsClass )
-		out << "	new(input->user.token) " << Token() << "(*token);\n";
+		out << "	new(queue->user.token) " << Token() << "(*token);\n";
 	else
-		out << "	input->user.token = *token;\n";
+		out << "	queue->user.token = *token;\n";
 
 	out <<
-		"	input->next = 0;\n"
-		"	input->retry = 0;\n"
-		"	input->child = 0;\n"
+		"	queue->next = 0;\n"
+		"	queue->retry = 0;\n"
+		"	queue->child = 0;\n"
+		"	queue->causeReduce = 0;\n"
 		"\n"
 		"again:\n"
-		"	if ( input == 0 )\n"
-		"		goto _out;\n"
-		"\n";
+		"	if ( input == 0 ) {\n"
+		"		if ( queue == 0 )\n"
+		"			goto _out;\n"
+		"\n"
+		"		input = queue;\n"
+		"		queue = queue->next;\n"
+		"		input->next = 0;\n";
 
 	if ( translateBlock != 0 )
 		writeInlineBlock( 0, translateBlock );
+	
+	out <<
+		"	}\n"
+		"\n";
 
 	out <<
 		"	lel = input;\n"
@@ -743,7 +764,7 @@ void ParseData::writeExec()
 		"		action += (lel->retry & " << lower << ");\n"
 		"\n"
 		"	if ( *action & " << act_sb << " ) {\n"
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << \"shifted: \" << " << lelNames() << "[lel->type];\n"
 		"		#endif\n"
 		"		input = input->next;\n"
@@ -756,11 +777,11 @@ void ParseData::writeExec()
 		"		else {\n"
 		"			lel->retry += 1;\n"
 		"			" << numRetry() << " += 1;\n"
-		"			#ifdef LOG_ACTIONS\n"
+		"			#ifdef KELBT_LOG_ACTIONS\n"
 		"			cerr << \" retry: \" << " << stackTop() << ";\n"
 		"			#endif\n"
 		"		}\n"
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << endl;\n"
 		"		#endif\n"
 		"	}\n"
@@ -768,7 +789,7 @@ void ParseData::writeExec()
 		"	if ( " << commitLen() << "[pos] != 0 ) {\n"
 		"		struct " << LangEl() << " *commitHead = " << stackTop() << ", *lel;\n"
 		"		int sp = 0, doExec = 0;\n"
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << \"commit encountered, executing final actions\" << endl;\n"
 		"		#endif\n"
 		"		if ( " << commitLen() << "[pos] < 0 )\n"
@@ -876,30 +897,18 @@ void ParseData::writeExec()
 		"	if ( *action & " << act_rb << " ) {\n"
 		"		int reduction = *action >> 2;\n"
 		"		struct " << LangEl() << " *redLel;\n"
-		"		if ( " << pool() << " == 0 ) {\n"
-		"			if ( " << freshPos() << " == " << FRESH_BLOCK << " ) {\n"
-		"				struct " << Block() << "* newBlock  = (struct " << Block() << "*) "
-								"malloc( sizeof(struct " << Block() << ") );\n"
-		"				newBlock->next = " << block() << ";\n"
-		"				" << block() << " = newBlock;\n"
-		"				" << freshEl() << " = newBlock->data;\n"
-		"				#ifdef LOG_ACTIONS\n"
-		"				cerr << \"allocating " << FRESH_BLOCK << " LangEls\" << endl;\n"
-		"				#endif\n"
-		"				" << freshPos() << " = 0;\n"
-		"			}\n"
-		"			redLel = " << freshEl() << " + " << freshPos() << "++;\n"
-		"		}\n"
-		"		else {\n"
-		"			redLel = " << pool() << ";\n"
-		"			" << pool() << " = " << pool() << "->next;\n"
-		"		}\n"
-		"		" << numNodes() << " += 1;\n"
+		"\n"
+		"		if ( input != 0 )\n"
+		"			input->causeReduce += 1;\n"
+		"\n";
+		ALLOCATE_EL("redLel") <<
+		"\n"
 		"		redLel->type = " << prodLhsIds() << "[reduction];\n"
 		"		redLel->reduction = reduction;\n"
 		"		redLel->child = 0;\n"
 		"		redLel->next = 0;\n"
 		"		redLel->retry = (lel->retry << 16);\n"
+		"		redLel->causeReduce = 0;\n"
 		"		lel->retry &= " << upper << ";\n"
 		"\n"
 		"		rhsLen = " << prodLengths() << "[reduction];\n"
@@ -917,7 +926,7 @@ void ParseData::writeExec()
 	writeReduceBlocks();
 
 	out <<
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << \"reduced: \"\n"
 		"				<< " << prodNames() << "[reduction]\n"
 		"				<< \" rhsLen: \" << rhsLen;\n"
@@ -927,12 +936,12 @@ void ParseData::writeExec()
 		"		else {\n"
 		"			redLel->retry += 0x10000;\n"
 		"			" << numRetry() << " += 1;\n"
-		"			#ifdef LOG_ACTIONS\n"
+		"			#ifdef KELBT_LOG_ACTIONS\n"
 		"			cerr << \" retry: \" << redLel;\n"
 		"			#endif\n"
 		"		}\n"
 		"\n"
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << endl;\n"
 		"		#endif\n"
 		"\n"
@@ -949,7 +958,7 @@ void ParseData::writeExec()
 		"		}\n"
 		"\n"
 		"		if ( induceReject ) {\n"
-		"			#ifdef LOG_ACTIONS\n"
+		"			#ifdef KELBT_LOG_ACTIONS\n"
 		"			cerr << \"error induced during reduction of \" <<\n"
 		"					" << lelNames() << "[redLel->type] << endl;\n"
 		"			#endif\n"
@@ -970,33 +979,32 @@ void ParseData::writeExec()
 		"	goto again;\n"
 		"\n"
 		"parseError:\n"
-		"	#ifdef LOG_BACKTRACK\n"
+		"	#ifdef KELBT_LOG_BACKTRACK\n"
 		"	cerr << \"hit error\" << endl;\n"
 		"	#endif\n";
 
-	if ( undoTransBlock != 0 )
-		writeInlineBlock( 0, undoTransBlock );
-
 	out << 
 		"	if ( " << numRetry() << " > 0 ) {\n"
+		"		struct " << LangEl() << " *redLel;\n"
+		"\n"
+		"		if ( input != 0 ) {\n"
+		"			redLel = input;\n"
+		"			goto have_undo_element;\n"
+		"		}\n"
+		"\n"
 		"		while ( 1 ) {\n"
-		"			struct " << LangEl() << " *redLel = " << stackTop() << ";\n"
+		"			redLel = " << stackTop() << ";\n"
 		"			if ( " << stackTop() << "->type < " << firstNonTermId << " ) {\n"
-		"				#ifdef LOG_BACKTRACK\n"
+		"				#ifdef KELBT_LOG_BACKTRACK\n"
 		"				cerr << \"backing up over terminal: \" <<\n"
 		"						" << lelNames() << "[" << stackTop() << "->type] << endl;\n"
 		"				#endif\n"
 		"				" << stackTop() << " = " << stackTop() << "->next;\n"
 		"				redLel->next = input;\n"
-		"				input = redLel;\n";
-
-	if ( undoTransBlock != 0 )
-		writeInlineBlock( 0, undoTransBlock );
-
-	out <<
+		"				input = redLel;\n"
 		"			}\n"
 		"			else {\n"
-		"				#ifdef LOG_BACKTRACK\n"
+		"				#ifdef KELBT_LOG_BACKTRACK\n"
 		"				cerr << \"backing up over non-terminal: \" <<\n"
 		"						" << lelNames() << "[" << stackTop() << "->type] << endl;\n"
 		"				#endif\n"
@@ -1027,14 +1035,34 @@ void ParseData::writeExec()
 		"				redLel->next = " << pool() << ";\n"
 		"				" << pool() << " = redLel;\n"
 		"				" << numNodes() << " -= 1;\n"
+		"\n"
+		"				if ( input != 0 )\n"
+		"					input->causeReduce -= 1;\n"
 		"			}\n"
 		"\n"
-		"			if ( redLel->retry > 0 ) {\n"
-		"				#ifdef LOG_BACKTRACK\n"
+		"have_undo_element:\n"
+		"			if ( redLel->retry == 0 ) {\n"
+		"				if ( input != 0 && input->causeReduce == 0 ) {\n"
+		"					#ifdef KELBT_LOG_BACKTRACK\n"
+		"					cerr << \"pushing back: \" << " << lelNames() <<
+								"[input->type] << endl;\n"
+		"					#endif\n";
+
+	if ( undoTransBlock != 0 )
+		writeInlineBlock( 0, undoTransBlock );
+
+	out <<
+		"					input->next = queue;\n"
+		"					queue = input;\n"
+		"					input = 0;\n"
+		"				}\n"
+		"			}\n"
+		"			else {\n"
+		"				#ifdef KELBT_LOG_BACKTRACK\n"
 		"				cerr << \"found retry targ: \" << redLel << endl;\n"
 		"				#endif\n"
 		"				" << numRetry() << " -= 1;\n"
-		"				#ifdef LOG_BACKTRACK\n"
+		"				#ifdef KELBT_LOG_BACKTRACK\n"
 		"				cerr << \"found retry: \" << redLel << endl;\n"
 		"				#endif\n"
 		"				if ( redLel->retry & " << lower << " )\n"
@@ -1067,7 +1095,7 @@ void ParseData::writeFinish()
 		"	struct " << Block() << " *todel = 0;\n"
 		"	struct " << LangEl() << " *lel = " << stackTop() << ";\n"
 		"	int sp = 0;\n"
-		"	#ifdef LOG_ACTIONS\n"
+		"	#ifdef KELBT_LOG_ACTIONS\n"
 		"	cerr << \"cleaning up stack\" << endl;\n"
 		"	#endif\n"
 		"\n"
@@ -1104,7 +1132,7 @@ void ParseData::writeFinish()
 	
 	out <<
 		"		}\n"
-		"		#ifdef LOG_ACTIONS\n"
+		"		#ifdef KELBT_LOG_ACTIONS\n"
 		"		cerr << \"deleted: \" << " << lelNames() << "[lel->type] << endl;\n"
 		"		#endif\n"
 		"	}\n"
